from animation import AnimatedBlinkable
from game import g
from locals import *
from scrollable_sprite import ScrollableSprite

class Character(ScrollableSprite, AnimatedBlinkable):
    INITIAL_JUMP_VELOC = 800    # pixels per second
    GRAVITY_ACCEL = 1200        # pixels per second^2

    ANIMATION_IDLE = None
    ANIMATION_JUMP = None

    def __init__(self, initial_x, initial_y, initial_facing):
        ScrollableSprite.__init__(self)
        AnimatedBlinkable.__init__(self)

        self.move_allowed = True
        self.can_fall = True

        self.left_pressed = False
        self.right_pressed = False
        self.dir_facing = initial_facing
        self.is_moving = False

        self.walk_speed_tweak = 0
        self.walk_accel_tweak = 0

        self.horiz_accel = 0
        self.horiz_veloc = 0

        self.vert_accel = 0
        self.vert_veloc = 0
        self.falling = False

        self.triggers_weak_platforms = False

        self.lowgrav = False

        self.walk_right_anim = g.assets[self.ANIMATION[DIR_RIGHT]]
        self.walk_left_anim = g.assets[self.ANIMATION[DIR_LEFT]]

        if self.ANIMATION_IDLE is None:
            self.idle_right_anim = self.walk_right_anim.get_frame(0)
            self.idle_left_anim = self.walk_left_anim.get_frame(0)
        else:
            self.idle_right_anim = g.assets[self.ANIMATION_IDLE[DIR_RIGHT]]
            self.idle_left_anim = g.assets[self.ANIMATION_IDLE[DIR_LEFT]]

        if self.ANIMATION_JUMP is None:
            self.jump_left_anim = self.idle_left_anim
            self.jump_right_anim = self.idle_right_anim
        else:
            self.jump_left_anim = g.assets[self.ANIMATION_JUMP[DIR_LEFT]]
            self.jump_right_anim = g.assets[self.ANIMATION_JUMP[DIR_RIGHT]]

        self._stand_anim()

        self.disp_rect = self.image.get_rect()
        self.pos = initial_x, initial_y

    @property
    def char_rect(self):
        return self.animation.char_rect.move(*self.disp_rect.topleft)

    @property
    def hurt_rect(self):
        return self.char_rect.inflate(-5, -5)

    @property
    def initial_jump_veloc(self):
        if self.lowgrav:
            return self.INITIAL_JUMP_VELOC
        else:
            return self.INITIAL_JUMP_VELOC

    @property
    def gravity_accel(self):
        if self.lowgrav:
            return int(self.GRAVITY_ACCEL * 0.6)
        else:
            return self.GRAVITY_ACCEL

    @property
    def initial_walk_veloc(self):
        if self.lowgrav:
            return (self.INITIAL_WALK_VELOC + self.walk_speed_tweak) / 2
        else:
            return self.INITIAL_WALK_VELOC + self.walk_speed_tweak

    @property
    def walk_accel(self):
        if self.lowgrav:
            return (self.WALK_ACCEL + self.walk_accel_tweak) / 2
        else:
            return self.WALK_ACCEL + self.walk_accel_tweak

    @property
    def walk_accel_duration(self):
        if self.lowgrav:
            return self.WALK_ACCEL_DURATION
        else:
            return self.WALK_ACCEL_DURATION

    @property
    def pos(self):
        return self.fpos

    @pos.setter
    def pos(self, value):
        x, y = value
        self.fpos = float(x), float(y)

        # I want to set self.char_rect.midbottom but I need to set self.disp_rect.midbottom...
        # Quick hack does this
        r = Rect(self.char_rect)
        d = self.disp_rect
        dx = r.left - d.left
        dy = r.top - d.top
        r.midbottom = x, y
        d.left = r.left - dx
        d.top = r.top - dy

    @property
    def falling_down(self):
        return self.falling and self.vert_veloc > 0

    def update(self, *args):
        self.update_position()
        ScrollableSprite.update(self, *args)
        AnimatedBlinkable.update(self, *args)

    def start_walk_right(self):
        self.right_pressed = True
        if not self.is_moving or self.dir_facing != DIR_RIGHT:
            self.dir_facing = DIR_RIGHT
            self.start_walk()

    def start_walk_left(self):
        self.left_pressed = True
        if not self.is_moving or self.dir_facing != DIR_LEFT:
            self.dir_facing = DIR_LEFT
            self.start_walk()

    def stop_walk_right(self):
        self.right_pressed = False
        if self.is_moving and self.dir_facing == DIR_RIGHT:
            if self.left_pressed:
                self.dir_facing = DIR_LEFT
                self.start_walk()
            else:
                self.stop_walk()

    def stop_walk_left(self):
        self.left_pressed = False
        if self.is_moving and self.dir_facing == DIR_LEFT:
            if self.right_pressed:
                self.dir_facing = DIR_RIGHT
                self.start_walk()
            else:
                self.stop_walk()

    def start_walk(self):
        if not self.move_allowed:
            return

        self.is_moving = True
        self.walk_start = g.get_ticks() / 1000. # in seconds...
        self.accel_time_remaining = self.walk_accel_duration
        if self.dir_facing == DIR_RIGHT:
            self.horiz_veloc = self.initial_walk_veloc
            self.horiz_accel = self.walk_accel
        elif self.dir_facing == DIR_LEFT:
            self.horiz_veloc = -self.initial_walk_veloc
            self.horiz_accel = -self.walk_accel

        self.update_anim()

    def stop_walk(self):
        self.horiz_accel = 0
        self.horiz_veloc = 0
        self.is_moving = False
        self.update_anim()

    def _stand_anim(self):
        if self.dir_facing == DIR_LEFT:
            self.animation = self.idle_left_anim
        else:
            self.animation = self.idle_right_anim

    def _jump_anim(self):
        if self.dir_facing == DIR_LEFT:
            self.animation = self.jump_left_anim
        else:
            self.animation = self.jump_right_anim

    def _walk_anim(self):
        if self.dir_facing == DIR_LEFT:
            self.animation = self.walk_left_anim
        else:
            self.animation = self.walk_right_anim

    def update_anim(self):
        if self.falling:
            self._jump_anim()
        elif self.is_moving:
            self._walk_anim()
        else:
            self._stand_anim()

        self.pos = self.pos # update display position due to change of animation

    def jump(self):
        if self.move_allowed and not self.falling:
            self.start_falling()
            self.vert_veloc = -self.initial_jump_veloc

    def start_falling(self):
        if not self.can_fall:
            return

        self.falling = True
        self.fall_start = g.get_ticks() / 1000. # in seconds...
        self.vert_accel = self.gravity_accel
        self.vert_veloc = 0
        self.update_anim()

    def stop_falling(self):
        self.falling = False
        self.update_anim()

    def update_position(self):
        t = g.get_ticks() / 1000. # in seconds...

        cur_x, cur_y = final_x, final_y = self.pos

        # Are we moving horizontally?
        if self.is_moving:
            dt = t - self.walk_start
            self.walk_start = t # for next time

            if self.horiz_accel != 0 and dt > self.accel_time_remaining:
                # accelerate with dt = accel_time_remaining
                dt2 = self.accel_time_remaining
                # dx = v0 dt + (1/2) a dt dt
                # dv = a dt
                final_x += dt2 * (self.horiz_veloc + self.horiz_accel * dt2 / 2.)
                self.horiz_veloc += dt2 * self.horiz_accel

                # continue calculating without this period
                dt -= dt2

                # no more acceleration
                self.accel_time_remaining = 0
                self.horiz_accel = 0

            # dx = v0 dt + (1/2) a dt dt
            # dv = a dt
            final_x += dt * (self.horiz_veloc + self.horiz_accel * dt / 2.)
            self.horiz_veloc += dt * self.horiz_accel

            if self.accel_time_remaining > 0:
                self.accel_time_remaining -= dt

        # Are we free-falling? (y acceleration)
        if self.falling:
            dt = t - self.fall_start
            self.fall_start = t # for next time

            # dy = v0 dt + (1/2) a dt dt
            # dv = a dt
            final_y += dt * (self.vert_veloc + self.vert_accel * dt / 2.)
            self.vert_veloc += dt * self.vert_accel

        while cur_y != final_y:
            if abs(cur_y - final_y) > 3:
                d = 5
            else:
                d = 1

            prev_y = cur_y
            prev_rect = Rect(self.char_rect)

            if cur_y > final_y:
                cur_y = max(cur_y - d, final_y)
            else:
                cur_y = min(cur_y + d, final_y)

            self.pos = cur_x, cur_y

            # Did we run into a solid platform?
            colliding_platforms = [plat for plat in g.level.solid_platforms
                                   if self.char_rect.colliderect(plat.disp_rect)]
            while colliding_platforms:
                plat = colliding_platforms.pop()
                plat_rect = plat.disp_rect

                # Did we hit it from the top or bottom?
                if prev_rect.bottom <= plat_rect.top < self.char_rect.bottom:
                    # From the top (fell on this platform) - don't care if it's walk-through, solid is enough to stop falling
                    self.stop_falling()
                    final_y = cur_y = prev_y
                    self.pos = cur_x, cur_y

                elif self.char_rect.top < plat_rect.bottom <= prev_rect.top:
                    # From the bottom (jumped into this platform)
                    if not plat.walk_through:
                        self.vert_veloc = 0 # "stop in midair"
                        final_y = cur_y = prev_y
                        self.pos = cur_x, cur_y

        while cur_x != final_x:
            if abs(cur_x - final_x) > 3:
                d = 5
            else:
                d = 1

            prev_x = cur_x
            prev_rect = Rect(self.char_rect)
            if cur_x > final_x:
                cur_x = max(cur_x - d, final_x)
            else:
                cur_x = min(cur_x + d, final_x)

            self.pos = cur_x, cur_y

            # Did we run into a solid platform?
            colliding_platforms = [plat for plat in g.level.solid_platforms
                                   if self.char_rect.colliderect(plat.disp_rect)]
            while colliding_platforms:
                plat = colliding_platforms.pop()
                plat_rect = plat.disp_rect

                # Did we hit it from the right or left?
                if self.char_rect.left < plat_rect.right <= prev_rect.left:
                    # From the right
                    if not plat.walk_through:
                        final_x = cur_x = prev_x
                        self.pos = cur_x, cur_y

                        if self.dir_facing == DIR_LEFT:
                            self.walked_into_wall()

                elif prev_rect.right <= plat_rect.left < self.char_rect.right:
                    # From the left
                    if not plat.walk_through:
                        final_x = cur_x = prev_x
                        self.pos = cur_x, cur_y

                        if self.dir_facing == DIR_RIGHT:
                            self.walked_into_wall()

        # Fall-related checks:
        if self.can_fall and not self.falling:
            solid_plat_rects = [plat.disp_rect for plat in g.level.solid_platforms]

            # Warn that we're about to fall.
            if self.is_moving and not self.falling:
                if self.dir_facing == DIR_RIGHT:
                    pt_x, pt_y = (self.char_rect.right - 5, self.char_rect.bottom + 2)
                else:
                    pt_x, pt_y = (self.char_rect.left + 5, self.char_rect.bottom + 2)

                idx = Rect(pt_x, pt_y, 1, 1).collidelist(solid_plat_rects)
                if idx == -1:
                    self.about_to_fall()

            # Check out whether we need to fall. If we don't have a solid platform underneath, fall
            underline = Rect(self.char_rect.left, self.char_rect.bottom + 1, self.char_rect.width, 1)
            indexes = underline.collidelistall(solid_plat_rects)

            if not indexes:
                # Nothing solid underneath
                self.start_falling()
            elif self.triggers_weak_platforms:
                # Trigger all weak platforms we're stepping on
                for idx in indexes:
                    g.level.solid_platforms[idx].step()

    def walked_into_wall(self):
        pass

    def about_to_fall(self):
        pass
