import random
from background import Background
from enemy import RedEnemy, GreenEnemy
from game import g
from locals import *
from platform import Floor, RegularPlatform, Block, PitRight, PitLeft, FloorTop, BlockTop, RegularWeakPlatform, InvisibleWall, EndSign, FogOverlay, WaterOverlay, AlternatePlatform, Stalactite1, Stalactite2, Stalagmite, InvisibleWallRight
from player import Player
from powerup import MedKit, ExtraLife
from scorebar import ScoreBar

####################################################################################################

class Level(object):
    HORIZONTAL = True
    DARK = None
    SUPERSCROLLING_SPEED = 0
    LOW_GRAVITY = False

    SCROLL_THRESHOLD_W = int(0.6 * SCREEN_WIDTH) # % of screen width that the character reaches before the level scrolls
    SCROLL_THRESHOLD_H = int(0.7 * SCREEN_HEIGHT)
    SPAWN_DISTANCE = int(1.1 * SCREEN_WIDTH) # how far away we should be before the enemy spawns

    def __init__(self):
        self.scroll_offset = 0
        self.cheat = 0
        self.superscroll_t = g.get_ticks()

        # Create sprite groups
        self.background = pygame.sprite.RenderUpdates()
        self.characters = pygame.sprite.RenderUpdates()
        self.bullets = pygame.sprite.RenderUpdates()
        self.platforms = pygame.sprite.OrderedUpdates()
        self.solid_platforms = []
        self.scorebar = ScoreBar()
        self.powerups = pygame.sprite.RenderUpdates()
        self.foreground = pygame.sprite.OrderedUpdates()

        # Load the level
        self._load_level()

    def _load_level(self):
        # Get the level asset
        level_asset = g.assets[self.LEVEL_ASSET]
        if self.HORIZONTAL:
            self.width, self.height = level_asset.width, SCREEN_HEIGHT
        else:
            self.width, self.height = SCREEN_WIDTH, level_asset.height

        # Load background
        self.background.add(Background())

        # Create player
        x, y = level_asset.player[0]
        self.player = Player(x, y)
        self.player.lowgrav = self.LOW_GRAVITY
        self.characters.add(self.player)

        # Create enemy spawn points
        self.enemies = []
        #self.enemy1_spawn = sorted(level_asset.enemy1, key = lambda pos : pos[0], reverse = True) # LIFO
        #self.enemy2_spawn = sorted(level_asset.enemy2, key = lambda pos : pos[0], reverse = True) # LIFO
        # Note: I've given up on sorting, it's not that important
        self.enemy1_spawn = level_asset.enemy1[:]
        self.enemy2_spawn = level_asset.enemy2[:]

        # Create platforms
        self.add_platform(InvisibleWall(0, 0))
        if not self.HORIZONTAL:
            self.add_platform(InvisibleWallRight(self.width, 0))

        for x, y in level_asset.floor:
            self.add_platform(Floor(x, y))
            if g.assets.has_key(FloorTop.ASSET_NAME):
                self.foreground.add(FloorTop(x, y))

        for x, y in level_asset.platform:
            self.add_platform(RegularPlatform(x, y))

        for x, y in level_asset.platform2:
            self.add_platform(AlternatePlatform(x, y))

        for x, y in level_asset.weak_platform:
            self.add_platform(RegularWeakPlatform(x, y))

        for x, y in level_asset.block:
            self.add_platform(Block(x, y))
            if g.assets.has_key(BlockTop.ASSET_NAME):
                self.foreground.add(BlockTop(x, y))

        for x, y in level_asset.pit_left:
            self.add_platform(PitLeft(x, y))

        for x, y in level_asset.pit_right:
            self.add_platform(PitRight(x, y))

        for x, y in level_asset.end_sign:
            self.add_platform(EndSign(x, y))

        # Create power-ups
        for x, y in level_asset.pickup1 + level_asset.pickup2 + level_asset.pickup3 + level_asset.pickup4:
            self.powerups.add(MedKit(x, y))

        for x, y in level_asset.pickup5 + level_asset.pickup6:
            self.powerups.add(ExtraLife(x, y))

        for x, y in level_asset.weapon:
            self.powerups.add(self.LEVEL_WEAPON(x, y))

        # Load stalactite and stalagmite foreground objects
        for x, y in level_asset.stalactite1:
            self.foreground.add(Stalactite1(x, y))

        for x, y in level_asset.stalactite2:
            self.foreground.add(Stalactite2(x, y))

        for x, y in level_asset.stalagmite:
            self.foreground.add(Stalagmite(x, y))

        # Load fog over all the other foreground, if needed
        if self.DARK:
            if self.DARK == 'fog-overlay':
                self.foreground.add(FogOverlay(0, 0))
            elif self.DARK == 'water-overlay':
                self.foreground.add(WaterOverlay(0, 0))

    def add_platform(self, plat):
        self.platforms.add(plat)
        if plat.is_solid:
            self.solid_platforms.append(plat)

    def scroll(self):
        if self.SUPERSCROLLING_SPEED:
            return

        if self.HORIZONTAL:
            # Get the player's on-screen rect (the char_rect with the current scroll offset)
            player_rect = self.player.char_rect.move(self.scroll_offset, 0)

            offset = player_rect.right - self.SCROLL_THRESHOLD_W
            if offset >= 0:
                self.scroll_offset -= offset
                if self.scroll_offset < SCREEN_WIDTH - self.width:
                    self.scroll_offset = SCREEN_WIDTH - self.width

            offset = SCREEN_WIDTH - self.SCROLL_THRESHOLD_W - player_rect.left
            if offset >= 0:
                self.scroll_offset += offset
                if self.scroll_offset > 0:
                    self.scroll_offset = 0

        else:
            # Get the player's on-screen rect (the char_rect with the current scroll offset)
            player_rect = self.player.char_rect.move(0, self.scroll_offset)

            offset = player_rect.bottom - self.SCROLL_THRESHOLD_H
            if offset >= 0:
                self.scroll_offset -= offset
                if self.scroll_offset < SCREEN_HEIGHT - self.height:
                    self.scroll_offset = SCREEN_HEIGHT - self.height

            offset = SCREEN_HEIGHT - self.SCROLL_THRESHOLD_H - player_rect.top
            if offset >= 0:
                self.scroll_offset += offset
                if self.scroll_offset > 0:
                    self.scroll_offset = 0

    def play(self):
        if g.assets.has_key("level-voice"):
            snd = g.assets["level-voice"]["sound"]
            snd.play()

        g.score.restart_level(self.LEVEL_WEAPON)
        self.looping = True

        # First time, refresh entire screen
        self.refresh_entire_screen = True
        g.tick() # Baseline

        while self.looping:
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYUP and event.key == K_ESCAPE):
                    self.result = USER_EXIT
                    self.looping = False
                elif g.cheating and event.type == KEYDOWN and K_1 <= event.key <= K_9:
                    self.looping = False
                    self.result = CHEATED
                    g.ls.cheat(event.key - K_1)

                else:
                    # player events
                    self.player.handle_event(event)

                    # global events
                    if g.handle_event(event):
                        self.refresh_entire_screen = True

            # Spawn/unspawn enemies in time for their first update()
            self.enemy1_spawn = self.check_spawn(self.enemy1_spawn, GreenEnemy)
            self.enemy2_spawn = self.check_spawn(self.enemy2_spawn, RedEnemy)

            # Update all the sprites
            self.background.update(self.scroll_offset, self.HORIZONTAL)
            self.platforms.update(self.scroll_offset, self.HORIZONTAL)
            self.powerups.update(self.scroll_offset, self.HORIZONTAL)
            self.characters.update(self.scroll_offset, self.HORIZONTAL)
            self.bullets.update(self.scroll_offset, self.HORIZONTAL)
            self.foreground.update(self.scroll_offset, self.HORIZONTAL)
            self.scorebar.update(self.scroll_offset, self.HORIZONTAL)

            self.check_fall()
            self.check_win()
            self.check_stomp()
            self.check_melee()
            self.check_bullet()
            self.check_powerup()

            # Scroll the screen before drawing
            self.scroll()

            update_rects = []
            update_rects += self.background.draw(g.screen)
            update_rects += self.platforms.draw(g.screen)
            update_rects += self.powerups.draw(g.screen)
            update_rects += self.characters.draw(g.screen)
            update_rects += self.bullets.draw(g.screen)
            update_rects += self.foreground.draw(g.screen)
            update_rects += self.scorebar.draw(g.screen)

            if self.refresh_entire_screen:
                pygame.display.update()
                self.refresh_entire_screen = False
            else:
                pygame.display.update(update_rects)

            self.tick()
            g.tick()

    def spawn_enemy(self, x, y, enemy_class = None):
        if enemy_class is None:
            enemy_class = random.choice([RedEnemy, GreenEnemy])

        e = enemy_class(x, y)
        e.start_walk()
        self.enemies.append(e)
        self.characters.add(e)

    def check_spawn(self, spawn_stack, enemy_class):
        new_stack = []
        px, py = self.player.pos

        while spawn_stack:
            # Check whether this guy should be spawned
            x, y = spawn_stack.pop()
            dx = x - px
            dy = y - py
            if dx * dx + dy * dy <= self.SPAWN_DISTANCE * self.SPAWN_DISTANCE:
                if DEBUG_MODE:
                    print "Spawning %r at %d, %d (%d active enemies)" % (enemy_class, x, y, len(self.enemies))
                self.spawn_enemy(x, y, enemy_class)
            else:
                new_stack.append((x, y))

        # Remove far away enemies from the screen because it slows the game down a lot, esp. on level 2
        kill_list = []
        for enemy in self.enemies:
            if enemy.__class__ == enemy_class:
                x, y = enemy.pos
                dx = x - px
                dy = y - py
                if dx * dx + dy * dy > self.SPAWN_DISTANCE * self.SPAWN_DISTANCE:
                    if DEBUG_MODE:
                        print "Unspawning %r at %d, %d (%d active enemies)" % (enemy_class, x, y, len(self.enemies))
                    kill_list.append(enemy)
                    new_stack.append((x, y))

        while kill_list:
            self.kill_enemy_no_attack(kill_list.pop())

        return new_stack

    def check_fall(self):
        # Did player fall off?
        if self.player.char_rect.top > self.height:
            self.player_died()

    def check_stomp(self):
        me_rect = self.player.hurt_rect

        # Stomp/touched an enemy?
        stomp_success = False

        # Find all enemies we're touching
        indexes = me_rect.collidelistall([ e.char_rect for e in self.enemies ])
        kill_list = [self.enemies[idx] for idx in indexes]

        for e in kill_list:
            # Try stomping it... (Must be falling on it and it must be stomp-compatible)
            if (not e.immunities.has_immunity(ATTACK_STOMP)) and self.player.falling_down and self.player.char_rect.bottom < e.hurt_rect.centery:
                self.kill_enemy(e, ATTACK_STOMP)
                stomp_success = True
            else:
                if not self.player.hurt():
                    self.player_died()
                    return

        if stomp_success:
            self.player.stomp_jump()

    def check_melee(self):
        if self.player.in_melee and self.player.hit_rect:
            indexes = self.player.hit_rect.collidelistall([ e.hurt_rect for e in self.enemies ])
            kill_list = [self.enemies[idx] for idx in indexes]

            for e in kill_list:
                # Try attacking it... (Must be compatible with the attack)
                if not e.immunities.has_immunity(self.player.melee.ATTACK_TYPE):
                    self.kill_enemy(e, self.player.melee.ATTACK_TYPE)
                else:
                    if not self.player.hurt():
                        self.player_died()
                        return

    def check_bullet(self):
        bullets = list(self.bullets)
        for bullet in bullets:
            # Get first enemy hurt by this bullet
            idx = bullet.char_rect.collidelist([ e.hurt_rect for e in self.enemies ])
            if idx == -1:
                continue

            # An enemy was hit by this bullet!
            e = self.enemies[idx]

            # Kill the bullet and maybe the enemy
            bullet.hit_something()
            if not e.immunities.has_immunity(bullet.ATTACK_TYPE):
                self.kill_enemy(e, bullet.ATTACK_TYPE)

    def check_powerup(self):
        for p in self.powerups:
            if p.disp_rect.colliderect(self.player.char_rect):
                p.take()

    def kill_enemy_no_attack(self, enemy):
        enemy.kill()
        self.enemies.remove(enemy)

    def kill_enemy(self, enemy, attack):
        g.score.register_attack(enemy, attack)
        enemy.death_sound()
        self.kill_enemy_no_attack(enemy)

    def player_died(self):
        self.result = PLAYER_DEAD
        self.looping = False

    def check_win(self):
        if self.player.char_rect.left > self.width - 3:
            self.result = LEVEL_BEATEN
            self.looping = False

    def tick(self):
        pass